/*
	Copyright 2012 dataglot.org
	
	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at
   
		http://www.apache.org/licenses/LICENSE-2.0
	   
	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
	
	This file was changed by dataglot.org; the source copyright notice follows:
	
	Copyright 2006 The Apache Software Foundation.
	
	Licensed under the Apache License, Version 2.0 (the "License")
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at
	
		http://www.apache.org/licenses/LICENSE-2.0
		
	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
 */
package org.dataglot.jaxb.plugin.fluent;

import com.sun.codemodel.JClass;
import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.Plugin;
import com.sun.tools.xjc.outline.ClassOutline;
import com.sun.tools.xjc.outline.Outline;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.xml.sax.ErrorHandler;

public class FluentPlugin extends Plugin
{
	@Override
	public String getOptionName()
	{
		return "Xfluent-api";
	}

	@Override
	public String getUsage()
	{
		return "  -Xfluent-api		:  enable fluent api for generated code";
	}

	@Override
	public boolean run(final Outline outline, final Options options, final ErrorHandler errorHandler)
	{
		final JType voidType = outline.getCodeModel().VOID;
		for (ClassOutline classOutline : outline.getClasses())
		{
			final JDefinedClass targetImplClass = classOutline.implClass;
			final Collection<Method> methods = new ArrayList<Method>();
			final Set<String> methodNames = new HashSet<String>();
			boolean isOverride = false;

			for (;;)
			{
				final JDefinedClass implClass = classOutline.implClass;
				for (final JMethod method : implClass.methods())
				{
					if (methodNames.contains(method.name()))
					{
						continue;
					}
					if (isSetterMethod(method, voidType))
					{
						methods.add(new Method(method, MethodType.FLUENT_SETTER, isOverride));
						methodNames.add(method.name());
					}
					else if (isListGetterMethod(method))
					{
						methods.add(new Method(method, MethodType.FLUENT_LIST_SETTER, isOverride));
						methods.add(new Method(method, MethodType.FLUENT_COLLECTION_SETTER, isOverride));
						methodNames.add(method.name());
					}
				}
				classOutline = classOutline.getSuperClass();

				if (classOutline == null)
				{
					break;
				}
				isOverride = true;
			}
			for (final Method method : methods)
			{
				method.createMethodInClass(targetImplClass);
			}
		}
		return true;
	}

	private boolean isInt(final JType type)
	{
		final JCodeModel codeModel = type.owner();
		return type.isPrimitive() && codeModel.INT.equals(JType.parse(codeModel, type.name()));
	}

	private boolean isListGetterMethod(final JMethod jmethod)
	{
		final int mods = jmethod.mods().getValue();
		if ((mods & JMod.STATIC) == 1 || (mods & JMod.PUBLIC) == 0)
		{
			return false;
		}
		final String methodName = jmethod.name();
		if (methodName.length() <= MethodType.GETTER_METHOD_PREFIX_LEN || !methodName.startsWith(MethodType.GETTER_METHOD_PREFIX))
		{
			return false;
		}
		// A list getter method will have no argument.
		if (jmethod.listParams().length > 0)
		{
			return false;
		}
		// See if the return type of the method is a List<T>.
		final JType jtype = jmethod.type();

		if (jtype instanceof JClass)
		{
			final JClass jclass = JClass.class.cast(jtype);
			final List<JClass> typeParams = jclass.getTypeParameters();

			if (typeParams.size() != 1)
			{
				return false;
			}
			return jclass.fullName().startsWith(MethodType.PARAMETERIZED_LIST_PREFIX);
		}
		return false;
	}

	private boolean isSetterMethod(final JMethod method, final JType voidType)
	{
		// Return type of a setter method is expected to be void.
		if (method.type() == voidType)
		{
			final JVar[] params = method.listParams();

			switch (params.length)
			{
			case 2:
				// Could be an indexed property setter method.
				// If so, the first argument must be the index (a primitive int).
				if (!isInt(params[0].type()))
				{
					return false;
				}
				// Fall through.
			case 1:
				// Or could be a simple property setter method.
				final int mods = method.mods().getValue();

				if ((mods & JMod.STATIC) == 0 && (mods & JMod.PUBLIC) == 1)
				{
					final String methodName = method.name();
					return methodName.length() > MethodType.SETTER_METHOD_PREFIX_LEN && methodName.startsWith(MethodType.SETTER_METHOD_PREFIX);
				}
				break;
			}
		}
		return false;
	}
}
