/*
	Copyright 2012 dataglot.org
	
	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at
   
		http://www.apache.org/licenses/LICENSE-2.0
	   
	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
	
	This file was changed by dataglot.org; the source copyright notice follows:
	
	Copyright 2006 The Apache Software Foundation.
	
	Licensed under the Apache License, Version 2.0 (the "License")
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at
	
		http://www.apache.org/licenses/LICENSE-2.0
		
	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
 */
package org.dataglot.jaxb.plugin.fluent;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JClass;
import com.sun.codemodel.JConditional;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JForEach;
import com.sun.codemodel.JInvocation;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

public enum MethodType
{
	FLUENT_COLLECTION_SETTER {
		@Override
		public void createFluentMethod(final JDefinedClass implClass, final Method method)
		{
			final JMethod listGetterMethod = method.originalMethod();
			final String name = listGetterMethod.name();
			final int mods = JMod.PUBLIC | listGetterMethod.mods().getValue() & JMod.FINAL;
			final JMethod implMethod = implClass.method(mods, implClass, methodName(name.substring(GETTER_METHOD_PREFIX_LEN)));
			if (method.overridesParentMethod())
			{
				implMethod.annotate(Override.class);
			}
			final JType returnJType = listGetterMethod.type();
			// As is already checked in isListGetterMethod(JMethod):
			// 1) the return type must be a subtype of JClass; and
			// 2) the number of type parameters must be 1
			final JClass returnJClass = JClass.class.cast(returnJType);
			final List<JClass> typeParams = returnJClass.getTypeParameters();
			assert typeParams.size() == 1;
			final JClass typeParam = typeParams.get(0);
			// Support Collection with type parameter
			final JClass narrowedCollectionJClass = implClass.owner().ref(Collection.class).narrow(typeParam);
			final JVar jvarParam = implMethod.param(narrowedCollectionJClass, VALUES);
			final JBlock body = implMethod.body();
			final JConditional cond = body._if(jvarParam.ne(JExpr._null()));
			final JInvocation addInvocation = cond._then().invoke(JExpr.invoke(listGetterMethod), "addAll");
			addInvocation.arg(jvarParam);
			body._return(JExpr._this());
			return;
		}
	},
	FLUENT_LIST_SETTER {
		@Override
		public void createFluentMethod(final JDefinedClass implClass, final Method method)
		{
			final JMethod listGetterMethod = method.originalMethod();
			final String name = listGetterMethod.name();
			final int mods = JMod.PUBLIC | listGetterMethod.mods().getValue() & JMod.FINAL;
			final JMethod implMethod = implClass.method(mods, implClass, methodName(name.substring(GETTER_METHOD_PREFIX_LEN)));
			if (method.overridesParentMethod())
			{
				implMethod.annotate(Override.class);
			}
			final JType returnJType = listGetterMethod.type();
			// As is already checked in isListGetterMethod(JMethod):
			// 1) the return type must be a subtype of JClass; and
			// 2) the number of type parameters must be 1
			final JClass returnJClass = JClass.class.cast(returnJType);
			final List<JClass> typeParams = returnJClass.getTypeParameters();
			assert typeParams.size() == 1;
			final JClass typeParam = typeParams.get(0);
			// Support variable arguments
			final JVar jvarParam = implMethod.varParam(typeParam, VALUES);
			final JBlock body = implMethod.body();
			final JConditional cond = body._if(jvarParam.ne(JExpr._null()));
			final JForEach forEach = cond._then().forEach(typeParam, VALUE, JExpr.ref(VALUES));
			final JInvocation addInvocation = forEach.body().invoke(JExpr.invoke(listGetterMethod), "add");
			addInvocation.arg(JExpr.ref(VALUE));
			body._return(JExpr._this());
			return;
		}
	},
	FLUENT_SETTER {
		@Override
		public void createFluentMethod(final JDefinedClass implClass, final Method method)
		{
			final JMethod original = method.originalMethod();
			final String name = original.name();
			final int mods = JMod.PUBLIC | original.mods().getValue() & JMod.FINAL;
			final JMethod implMethod = implClass.method(mods, implClass, methodName(name.substring(SETTER_METHOD_PREFIX_LEN)));
			if (method.overridesParentMethod())
			{
				implMethod.annotate(Override.class);
			}
			final JVar[] jvars = original.listParams();
			// jvars.length == 1 means simple property setter method.
			// jvars.length == 2 means indexed property setter method.
			assert jvars.length == 1 || jvars.length == 2;
			// with the same parameter(s) as the set* method
			for (final JVar jvar : jvars)
			{
				implMethod.param(jvar.mods().getValue(), jvar.type(), jvar.name());
			}
			final JBlock jblock = implMethod.body();
			// The with* method in turn invoke the setter method
			final JInvocation jinvocation = jblock.invoke(original);
			// passing the same list of arguments
			for (final JVar jvar : jvars)
			{
				jinvocation.arg(jvar);
			}
			jblock._return(JExpr._this());
			return;
		}
	};
	
	public abstract void createFluentMethod(final JDefinedClass implClass, final Method method);

	private static String methodName(final String initialCaps)
	{
		final String firstLetter = initialCaps.substring(0, 1);
		final String remainder = initialCaps.substring(1);
		final String methodName = firstLetter.toLowerCase() + remainder;
		if (reserved.contains(methodName))
		{
			return methodName + "_";
		}
		else
		{
			return methodName;
		}
	}

	public static final String GETTER_METHOD_PREFIX = "get";

	public static final int GETTER_METHOD_PREFIX_LEN = GETTER_METHOD_PREFIX.length();

	public static final String PARAMETERIZED_LIST_PREFIX = List.class.getName() + "<";

	public static final String SETTER_METHOD_PREFIX = "set";

	public static final int SETTER_METHOD_PREFIX_LEN = SETTER_METHOD_PREFIX.length();

	private static final List<String> reserved = Arrays.asList("abstract", "assert", "boolean", "break", "byte", "case", "catch", "char", "char", "class", "const", "continue", "default", "do",
		"double", "else", "enum", "extends", "false", "final", "finally", "float", "for", "for", "goto", "if", "implements", "import", "instanceof", "int", "interface", "long", "native", "new",
		"null", "package", "private", "protected", "public", "return", "short", "static", "strictfp", "super", "switch", "synchronized", "this", "throw", "throws", "transient", "true", "try", "void",
		"volatile", "while");

	private static final String VALUE = "value";

	private static final String VALUES = "values";
}
